from semantic_model_generator.clickzetta_utils.clickzetta_connector import (
    ClickzettaConnectionProxy,
)


def load_yaml(yaml_path: str) -> str:
    """
    Load local yaml file into str.

    yaml_path: str The absolute path to the location of your yaml file. Something like path/to/your/file.yaml.
    """
    with open(yaml_path) as f:
        yaml_str = f.read()
    return yaml_str


def validate(yaml_str: str, conn: ClickzettaConnectionProxy) -> None:
    """
    Placeholder validation for ClickZetta. The ClickZetta API does not yet expose a
    dedicated semantic model validator, so we treat models as valid after basic checks.
    """
    _ = yaml_str
    _ = conn


def validate_from_local_path(yaml_path: str, conn: ClickzettaConnectionProxy) -> None:
    yaml_str = load_yaml(yaml_path)
    validate(yaml_str, conn)



import pytest

from semantic_model_generator.data_processing.data_types import FQNParts
from semantic_model_generator.clickzetta_utils.utils import create_fqn_table


def test_fqn_creation():
    input_name = "database.schema.table"

    fqn_parts = create_fqn_table(input_name)

    assert fqn_parts == FQNParts(
        database="DATABASE", schema_name="SCHEMA", table="table"
    )


def test_fqn_creation_invalid_name():
    input_name = "database.schema table"
    with pytest.raises(ValueError):
        create_fqn_table(input_name)



from semantic_model_generator import generate_model
from semantic_model_generator.data_processing.data_types import Column, Table


def test_identifier_like_numeric_stays_dimension() -> None:
    raw_table = Table(
        id_=0,
        name="CUSTOMERS",
        columns=[
            Column(id_=0, column_name="customer_id", column_type="INT"),
            Column(id_=1, column_name="loyalty_score", column_type="NUMBER"),
        ],
    )

    table = generate_model._raw_table_to_semantic_context_table(
        database="QUICK_START",
        schema="MCP_DEMO",
        raw_table=raw_table,
    )

    dimension_names = {dim.name for dim in table.dimensions}
    fact_names = {fact.name for fact in table.facts}

    assert "customer_id" in dimension_names
    assert "customer_id" not in fact_names
    assert "loyalty_score" in fact_names


def test_string_date_promoted_to_time_dimension() -> None:
    raw_table = Table(
        id_=0,
        name="ORDERS",
        columns=[
            Column(id_=0, column_name="order_date", column_type="STRING", values=["2024-01-01", "2024-02-01"]),
            Column(id_=1, column_name="order_status", column_type="STRING", values=["OPEN", "CLOSED"]),
        ],
    )

    table = generate_model._raw_table_to_semantic_context_table(
        database="QUICK_START",
        schema="MCP_DEMO",
        raw_table=raw_table,
    )

    time_dimension_names = {td.name for td in table.time_dimensions}
    dimension_names = {dim.name for dim in table.dimensions}

    assert "order_date" in time_dimension_names
    assert "order_date" not in dimension_names
    time_dim = next(td for td in table.time_dimensions if td.name == "order_date")
    assert time_dim.data_type == "TIMESTAMP_NTZ"


def test_identifier_sanitization_and_primary_key_detection() -> None:
    assert generate_model._sanitize_identifier_name("o_orderkey") == "ORDERKEY"
    assert generate_model._sanitize_identifier_name("L_PartKey") == "PARTKEY"
    assert generate_model._looks_like_primary_key("ORDERS", "o_orderkey")
    assert generate_model._looks_like_primary_key("NATION", "n_nationkey")



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AggregationType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    aggregation_type_unknown: _ClassVar[AggregationType]
    sum: _ClassVar[AggregationType]
    avg: _ClassVar[AggregationType]
    median: _ClassVar[AggregationType]
    min: _ClassVar[AggregationType]
    max: _ClassVar[AggregationType]
    count: _ClassVar[AggregationType]
    count_distinct: _ClassVar[AggregationType]

class ColumnKind(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    column_kind_unknown: _ClassVar[ColumnKind]
    dimension: _ClassVar[ColumnKind]
    measure: _ClassVar[ColumnKind]
    time_dimension: _ClassVar[ColumnKind]
    metric: _ClassVar[ColumnKind]

class JoinType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    join_type_unknown: _ClassVar[JoinType]
    inner: _ClassVar[JoinType]
    left_outer: _ClassVar[JoinType]
    full_outer: _ClassVar[JoinType]
    cross: _ClassVar[JoinType]
    right_outer: _ClassVar[JoinType]

class RelationshipType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    relationship_type_unknown: _ClassVar[RelationshipType]
    one_to_one: _ClassVar[RelationshipType]
    many_to_one: _ClassVar[RelationshipType]
    one_to_many: _ClassVar[RelationshipType]
    many_to_many: _ClassVar[RelationshipType]
aggregation_type_unknown: AggregationType
sum: AggregationType
avg: AggregationType
median: AggregationType
min: AggregationType
max: AggregationType
count: AggregationType
count_distinct: AggregationType
column_kind_unknown: ColumnKind
dimension: ColumnKind
measure: ColumnKind
time_dimension: ColumnKind
metric: ColumnKind
join_type_unknown: JoinType
inner: JoinType
left_outer: JoinType
full_outer: JoinType
cross: JoinType
right_outer: JoinType
relationship_type_unknown: RelationshipType
one_to_one: RelationshipType
many_to_one: RelationshipType
one_to_many: RelationshipType
many_to_many: RelationshipType
OPTIONAL_FIELD_NUMBER: _ClassVar[int]
optional: _descriptor.FieldDescriptor
SQL_EXPRESSION_FIELD_NUMBER: _ClassVar[int]
sql_expression: _descriptor.FieldDescriptor
ID_FIELD_FIELD_NUMBER: _ClassVar[int]
id_field: _descriptor.FieldDescriptor

class RetrievalResult(_message.Message):
    __slots__ = ("value", "score")
    VALUE_FIELD_NUMBER: _ClassVar[int]
    SCORE_FIELD_NUMBER: _ClassVar[int]
    value: str
    score: float
    def __init__(self, value: _Optional[str] = ..., score: _Optional[float] = ...) -> None: ...

class Column(_message.Message):
    __slots__ = ("name", "synonyms", "description", "expr", "data_type", "kind", "unique", "default_aggregation", "sample_values", "index_and_retrieve_values", "retrieved_literals", "cortex_search_service_name", "cortex_search_service", "is_enum")
    NAME_FIELD_NUMBER: _ClassVar[int]
    SYNONYMS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    EXPR_FIELD_NUMBER: _ClassVar[int]
    DATA_TYPE_FIELD_NUMBER: _ClassVar[int]
    KIND_FIELD_NUMBER: _ClassVar[int]
    UNIQUE_FIELD_NUMBER: _ClassVar[int]
    DEFAULT_AGGREGATION_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_VALUES_FIELD_NUMBER: _ClassVar[int]
    INDEX_AND_RETRIEVE_VALUES_FIELD_NUMBER: _ClassVar[int]
    RETRIEVED_LITERALS_FIELD_NUMBER: _ClassVar[int]
    CORTEX_SEARCH_SERVICE_NAME_FIELD_NUMBER: _ClassVar[int]
    CORTEX_SEARCH_SERVICE_FIELD_NUMBER: _ClassVar[int]
    IS_ENUM_FIELD_NUMBER: _ClassVar[int]
    name: str
    synonyms: _containers.RepeatedScalarFieldContainer[str]
    description: str
    expr: str
    data_type: str
    kind: ColumnKind
    unique: bool
    default_aggregation: AggregationType
    sample_values: _containers.RepeatedScalarFieldContainer[str]
    index_and_retrieve_values: bool
    retrieved_literals: _containers.RepeatedCompositeFieldContainer[RetrievalResult]
    cortex_search_service_name: str
    cortex_search_service: CortexSearchService
    is_enum: bool
    def __init__(self, name: _Optional[str] = ..., synonyms: _Optional[_Iterable[str]] = ..., description: _Optional[str] = ..., expr: _Optional[str] = ..., data_type: _Optional[str] = ..., kind: _Optional[_Union[ColumnKind, str]] = ..., unique: bool = ..., default_aggregation: _Optional[_Union[AggregationType, str]] = ..., sample_values: _Optional[_Iterable[str]] = ..., index_and_retrieve_values: bool = ..., retrieved_literals: _Optional[_Iterable[_Union[RetrievalResult, _Mapping]]] = ..., cortex_search_service_name: _Optional[str] = ..., cortex_search_service: _Optional[_Union[CortexSearchService, _Mapping]] = ..., is_enum: bool = ...) -> None: ...

class Dimension(_message.Message):
    __slots__ = ("name", "synonyms", "description", "expr", "data_type", "unique", "sample_values", "cortex_search_service", "cortex_search_service_name", "is_enum")
    NAME_FIELD_NUMBER: _ClassVar[int]
    SYNONYMS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    EXPR_FIELD_NUMBER: _ClassVar[int]
    DATA_TYPE_FIELD_NUMBER: _ClassVar[int]
    UNIQUE_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_VALUES_FIELD_NUMBER: _ClassVar[int]
    CORTEX_SEARCH_SERVICE_FIELD_NUMBER: _ClassVar[int]
    CORTEX_SEARCH_SERVICE_NAME_FIELD_NUMBER: _ClassVar[int]
    IS_ENUM_FIELD_NUMBER: _ClassVar[int]
    name: str
    synonyms: _containers.RepeatedScalarFieldContainer[str]
    description: str
    expr: str
    data_type: str
    unique: bool
    sample_values: _containers.RepeatedScalarFieldContainer[str]
    cortex_search_service: CortexSearchService
    cortex_search_service_name: str
    is_enum: bool
    def __init__(self, name: _Optional[str] = ..., synonyms: _Optional[_Iterable[str]] = ..., description: _Optional[str] = ..., expr: _Optional[str] = ..., data_type: _Optional[str] = ..., unique: bool = ..., sample_values: _Optional[_Iterable[str]] = ..., cortex_search_service: _Optional[_Union[CortexSearchService, _Mapping]] = ..., cortex_search_service_name: _Optional[str] = ..., is_enum: bool = ...) -> None: ...

class CortexSearchService(_message.Message):
    __slots__ = ("database", "schema", "service", "literal_column")
    DATABASE_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    SERVICE_FIELD_NUMBER: _ClassVar[int]
    LITERAL_COLUMN_FIELD_NUMBER: _ClassVar[int]
    database: str
    schema: str
    service: str
    literal_column: str
    def __init__(self, database: _Optional[str] = ..., schema: _Optional[str] = ..., service: _Optional[str] = ..., literal_column: _Optional[str] = ...) -> None: ...

class TimeDimension(_message.Message):
    __slots__ = ("name", "synonyms", "description", "expr", "data_type", "unique", "sample_values")
    NAME_FIELD_NUMBER: _ClassVar[int]
    SYNONYMS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    EXPR_FIELD_NUMBER: _ClassVar[int]
    DATA_TYPE_FIELD_NUMBER: _ClassVar[int]
    UNIQUE_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_VALUES_FIELD_NUMBER: _ClassVar[int]
    name: str
    synonyms: _containers.RepeatedScalarFieldContainer[str]
    description: str
    expr: str
    data_type: str
    unique: bool
    sample_values: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, name: _Optional[str] = ..., synonyms: _Optional[_Iterable[str]] = ..., description: _Optional[str] = ..., expr: _Optional[str] = ..., data_type: _Optional[str] = ..., unique: bool = ..., sample_values: _Optional[_Iterable[str]] = ...) -> None: ...

class Fact(_message.Message):
    __slots__ = ("name", "synonyms", "description", "expr", "data_type", "default_aggregation", "sample_values")
    NAME_FIELD_NUMBER: _ClassVar[int]
    SYNONYMS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    EXPR_FIELD_NUMBER: _ClassVar[int]
    DATA_TYPE_FIELD_NUMBER: _ClassVar[int]
    DEFAULT_AGGREGATION_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_VALUES_FIELD_NUMBER: _ClassVar[int]
    name: str
    synonyms: _containers.RepeatedScalarFieldContainer[str]
    description: str
    expr: str
    data_type: str
    default_aggregation: AggregationType
    sample_values: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, name: _Optional[str] = ..., synonyms: _Optional[_Iterable[str]] = ..., description: _Optional[str] = ..., expr: _Optional[str] = ..., data_type: _Optional[str] = ..., default_aggregation: _Optional[_Union[AggregationType, str]] = ..., sample_values: _Optional[_Iterable[str]] = ...) -> None: ...

class NamedFilter(_message.Message):
    __slots__ = ("name", "synonyms", "description", "expr")
    NAME_FIELD_NUMBER: _ClassVar[int]
    SYNONYMS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    EXPR_FIELD_NUMBER: _ClassVar[int]
    name: str
    synonyms: _containers.RepeatedScalarFieldContainer[str]
    description: str
    expr: str
    def __init__(self, name: _Optional[str] = ..., synonyms: _Optional[_Iterable[str]] = ..., description: _Optional[str] = ..., expr: _Optional[str] = ...) -> None: ...

class FullyQualifiedTable(_message.Message):
    __slots__ = ("database", "schema", "table")
    DATABASE_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    TABLE_FIELD_NUMBER: _ClassVar[int]
    database: str
    schema: str
    table: str
    def __init__(self, database: _Optional[str] = ..., schema: _Optional[str] = ..., table: _Optional[str] = ...) -> None: ...

class PrimaryKey(_message.Message):
    __slots__ = ("columns",)
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    columns: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, columns: _Optional[_Iterable[str]] = ...) -> None: ...

class ForeignKey(_message.Message):
    __slots__ = ("fkey_columns", "pkey_table", "pkey_columns")
    FKEY_COLUMNS_FIELD_NUMBER: _ClassVar[int]
    PKEY_TABLE_FIELD_NUMBER: _ClassVar[int]
    PKEY_COLUMNS_FIELD_NUMBER: _ClassVar[int]
    fkey_columns: _containers.RepeatedScalarFieldContainer[str]
    pkey_table: FullyQualifiedTable
    pkey_columns: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, fkey_columns: _Optional[_Iterable[str]] = ..., pkey_table: _Optional[_Union[FullyQualifiedTable, _Mapping]] = ..., pkey_columns: _Optional[_Iterable[str]] = ...) -> None: ...

class Table(_message.Message):
    __slots__ = ("name", "synonyms", "description", "base_table", "columns", "dimensions", "time_dimensions", "measures", "facts", "metrics", "primary_key", "foreign_keys", "filters")
    NAME_FIELD_NUMBER: _ClassVar[int]
    SYNONYMS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    BASE_TABLE_FIELD_NUMBER: _ClassVar[int]
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    DIMENSIONS_FIELD_NUMBER: _ClassVar[int]
    TIME_DIMENSIONS_FIELD_NUMBER: _ClassVar[int]
    MEASURES_FIELD_NUMBER: _ClassVar[int]
    FACTS_FIELD_NUMBER: _ClassVar[int]
    METRICS_FIELD_NUMBER: _ClassVar[int]
    PRIMARY_KEY_FIELD_NUMBER: _ClassVar[int]
    FOREIGN_KEYS_FIELD_NUMBER: _ClassVar[int]
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    name: str
    synonyms: _containers.RepeatedScalarFieldContainer[str]
    description: str
    base_table: FullyQualifiedTable
    columns: _containers.RepeatedCompositeFieldContainer[Column]
    dimensions: _containers.RepeatedCompositeFieldContainer[Dimension]
    time_dimensions: _containers.RepeatedCompositeFieldContainer[TimeDimension]
    measures: _containers.RepeatedCompositeFieldContainer[Fact]
    facts: _containers.RepeatedCompositeFieldContainer[Fact]
    metrics: _containers.RepeatedCompositeFieldContainer[Metric]
    primary_key: PrimaryKey
    foreign_keys: _containers.RepeatedCompositeFieldContainer[ForeignKey]
    filters: _containers.RepeatedCompositeFieldContainer[NamedFilter]
    def __init__(self, name: _Optional[str] = ..., synonyms: _Optional[_Iterable[str]] = ..., description: _Optional[str] = ..., base_table: _Optional[_Union[FullyQualifiedTable, _Mapping]] = ..., columns: _Optional[_Iterable[_Union[Column, _Mapping]]] = ..., dimensions: _Optional[_Iterable[_Union[Dimension, _Mapping]]] = ..., time_dimensions: _Optional[_Iterable[_Union[TimeDimension, _Mapping]]] = ..., measures: _Optional[_Iterable[_Union[Fact, _Mapping]]] = ..., facts: _Optional[_Iterable[_Union[Fact, _Mapping]]] = ..., metrics: _Optional[_Iterable[_Union[Metric, _Mapping]]] = ..., primary_key: _Optional[_Union[PrimaryKey, _Mapping]] = ..., foreign_keys: _Optional[_Iterable[_Union[ForeignKey, _Mapping]]] = ..., filters: _Optional[_Iterable[_Union[NamedFilter, _Mapping]]] = ...) -> None: ...

class Metric(_message.Message):
    __slots__ = ("name", "synonyms", "description", "expr", "filter")
    NAME_FIELD_NUMBER: _ClassVar[int]
    SYNONYMS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    EXPR_FIELD_NUMBER: _ClassVar[int]
    FILTER_FIELD_NUMBER: _ClassVar[int]
    name: str
    synonyms: _containers.RepeatedScalarFieldContainer[str]
    description: str
    expr: str
    filter: MetricsFilter
    def __init__(self, name: _Optional[str] = ..., synonyms: _Optional[_Iterable[str]] = ..., description: _Optional[str] = ..., expr: _Optional[str] = ..., filter: _Optional[_Union[MetricsFilter, _Mapping]] = ...) -> None: ...

class MetricsFilter(_message.Message):
    __slots__ = ("expr",)
    EXPR_FIELD_NUMBER: _ClassVar[int]
    expr: str
    def __init__(self, expr: _Optional[str] = ...) -> None: ...

class RelationKey(_message.Message):
    __slots__ = ("left_column", "right_column")
    LEFT_COLUMN_FIELD_NUMBER: _ClassVar[int]
    RIGHT_COLUMN_FIELD_NUMBER: _ClassVar[int]
    left_column: str
    right_column: str
    def __init__(self, left_column: _Optional[str] = ..., right_column: _Optional[str] = ...) -> None: ...

class Relationship(_message.Message):
    __slots__ = ("name", "left_table", "right_table", "expr", "relationship_columns", "join_type", "relationship_type")
    NAME_FIELD_NUMBER: _ClassVar[int]
    LEFT_TABLE_FIELD_NUMBER: _ClassVar[int]
    RIGHT_TABLE_FIELD_NUMBER: _ClassVar[int]
    EXPR_FIELD_NUMBER: _ClassVar[int]
    RELATIONSHIP_COLUMNS_FIELD_NUMBER: _ClassVar[int]
    JOIN_TYPE_FIELD_NUMBER: _ClassVar[int]
    RELATIONSHIP_TYPE_FIELD_NUMBER: _ClassVar[int]
    name: str
    left_table: str
    right_table: str
    expr: str
    relationship_columns: _containers.RepeatedCompositeFieldContainer[RelationKey]
    join_type: JoinType
    relationship_type: RelationshipType
    def __init__(self, name: _Optional[str] = ..., left_table: _Optional[str] = ..., right_table: _Optional[str] = ..., expr: _Optional[str] = ..., relationship_columns: _Optional[_Iterable[_Union[RelationKey, _Mapping]]] = ..., join_type: _Optional[_Union[JoinType, str]] = ..., relationship_type: _Optional[_Union[RelationshipType, str]] = ...) -> None: ...

class ModuleCustomInstructions(_message.Message):
    __slots__ = ("sql_generation", "question_categorization")
    SQL_GENERATION_FIELD_NUMBER: _ClassVar[int]
    QUESTION_CATEGORIZATION_FIELD_NUMBER: _ClassVar[int]
    sql_generation: str
    question_categorization: str
    def __init__(self, sql_generation: _Optional[str] = ..., question_categorization: _Optional[str] = ...) -> None: ...

class SemanticModel(_message.Message):
    __slots__ = ("name", "description", "tables", "metrics", "relationships", "verified_queries", "custom_instructions", "module_custom_instructions")
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    TABLES_FIELD_NUMBER: _ClassVar[int]
    METRICS_FIELD_NUMBER: _ClassVar[int]
    RELATIONSHIPS_FIELD_NUMBER: _ClassVar[int]
    VERIFIED_QUERIES_FIELD_NUMBER: _ClassVar[int]
    CUSTOM_INSTRUCTIONS_FIELD_NUMBER: _ClassVar[int]
    MODULE_CUSTOM_INSTRUCTIONS_FIELD_NUMBER: _ClassVar[int]
    name: str
    description: str
    tables: _containers.RepeatedCompositeFieldContainer[Table]
    metrics: _containers.RepeatedCompositeFieldContainer[Metric]
    relationships: _containers.RepeatedCompositeFieldContainer[Relationship]
    verified_queries: _containers.RepeatedCompositeFieldContainer[VerifiedQuery]
    custom_instructions: str
    module_custom_instructions: ModuleCustomInstructions
    def __init__(self, name: _Optional[str] = ..., description: _Optional[str] = ..., tables: _Optional[_Iterable[_Union[Table, _Mapping]]] = ..., metrics: _Optional[_Iterable[_Union[Metric, _Mapping]]] = ..., relationships: _Optional[_Iterable[_Union[Relationship, _Mapping]]] = ..., verified_queries: _Optional[_Iterable[_Union[VerifiedQuery, _Mapping]]] = ..., custom_instructions: _Optional[str] = ..., module_custom_instructions: _Optional[_Union[ModuleCustomInstructions, _Mapping]] = ...) -> None: ...

class VerifiedQuery(_message.Message):
    __slots__ = ("name", "semantic_model_name", "question", "sql", "verified_at", "verified_by", "use_as_onboarding_question")
    NAME_FIELD_NUMBER: _ClassVar[int]
    SEMANTIC_MODEL_NAME_FIELD_NUMBER: _ClassVar[int]
    QUESTION_FIELD_NUMBER: _ClassVar[int]
    SQL_FIELD_NUMBER: _ClassVar[int]
    VERIFIED_AT_FIELD_NUMBER: _ClassVar[int]
    VERIFIED_BY_FIELD_NUMBER: _ClassVar[int]
    USE_AS_ONBOARDING_QUESTION_FIELD_NUMBER: _ClassVar[int]
    name: str
    semantic_model_name: str
    question: str
    sql: str
    verified_at: int
    verified_by: str
    use_as_onboarding_question: bool
    def __init__(self, name: _Optional[str] = ..., semantic_model_name: _Optional[str] = ..., question: _Optional[str] = ..., sql: _Optional[str] = ..., verified_at: _Optional[int] = ..., verified_by: _Optional[str] = ..., use_as_onboarding_question: bool = ...) -> None: ...

class VerifiedQueryRepository(_message.Message):
    __slots__ = ("verified_queries",)
    VERIFIED_QUERIES_FIELD_NUMBER: _ClassVar[int]
    verified_queries: _containers.RepeatedCompositeFieldContainer[VerifiedQuery]
    def __init__(self, verified_queries: _Optional[_Iterable[_Union[VerifiedQuery, _Mapping]]] = ...) -> None: ...



from __future__ import annotations

from contextlib import contextmanager
from typing import Dict, Iterable

from clickzetta.zettapark.session import Session

from semantic_model_generator.data_processing.data_types import FQNParts

DEFAULT_HINTS: Dict[str, str] = {
    "sdk.job.timeout": "300",
    "query_tag": "semantic-model-generator",
    "cz.storage.parquet.vector.index.read.memory.cache": "true",
    "cz.storage.parquet.vector.index.read.local.cache": "false",
    "cz.sql.table.scan.push.down.filter": "true",
    "cz.sql.table.scan.enable.ensure.filter": "true",
    "cz.storage.always.prefetch.internal": "true",
    "cz.optimizer.generate.columns.always.valid": "true",
    "cz.sql.index.prewhere.enabled": "true",
    "cz.storage.parquet.enable.io.prefetch": "false",
}


def create_fqn_table(fqn_str: str) -> FQNParts:
    """
    Splits a fully qualified table name into its ClickZetta components.

    Expected format: ``{workspace}.{schema}.{table}``.
    """

    if fqn_str.count(".") != 2:
        raise ValueError(
            "Expected a fully-qualified identifier in the form "
            "{workspace}.{schema}.{table}; "
            f"received {fqn_str!r}"
        )
    workspace, schema, table = fqn_str.split(".")
    return FQNParts(
        database=workspace.upper(), schema_name=schema.upper(), table=table.upper()
    )


def _build_session_config(
    *,
    service: str,
    instance: str,
    workspace: str,
    schema: str,
    username: str,
    password: str,
    vcluster: str,
    hints: Dict[str, str] | None = None,
) -> Dict[str, object]:
    config: Dict[str, object] = {
        "service": service,
        "instance": instance,
        "workspace": workspace,
        "schema": schema,
        "username": username,
        "password": password,
        "vcluster": vcluster,
    }
    merged_hints = dict(DEFAULT_HINTS)
    if hints:
        merged_hints.update({k: str(v) for k, v in hints.items()})
    config["hints"] = merged_hints
    return config


def _apply_session_context(session: Session, *, schema: str, vcluster: str) -> None:
    for component, value in _iter_non_empty(
        ("schema", schema),
        ("vcluster", vcluster),
    ):
        session.sql(f"USE {component.upper()} {value.upper()}")


def _iter_non_empty(*pairs: tuple[str, str]) -> Iterable[tuple[str, str]]:
    for key, value in pairs:
        if value:
            yield key, value


def create_session(
    *,
    service: str,
    instance: str,
    workspace: str,
    schema: str,
    username: str,
    password: str,
    vcluster: str,
    hints: Dict[str, str] | None = None,
) -> Session:
    """
    Creates a ClickZetta Session pre-configured with workspace/schema context.
    """

    session = Session.builder.configs(
        _build_session_config(
            service=service,
            instance=instance,
            workspace=workspace,
            schema=schema,
            username=username,
            password=password,
            vcluster=vcluster,
            hints=hints,
        )
    ).create()
    _apply_session_context(session, schema=schema, vcluster=vcluster)
    return session


@contextmanager
def clickzetta_connection(
    *,
    service: str,
    instance: str,
    workspace: str,
    schema: str,
    username: str,
    password: str,
    vcluster: str,
    hints: Dict[str, str] | None = None,
) -> Session:
    """
    Context manager that yields a ClickZetta Session and ensures it is closed.
    """

    session = create_session(
        service=service,
        instance=instance,
        workspace=workspace,
        schema=schema,
        username=username,
        password=password,
        vcluster=vcluster,
        hints=hints,
    )
    try:
        yield session
    finally:
        session.close()



"""
Progress tracking system for semantic model enrichment process.
"""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Dict, Optional


class EnrichmentStage(str, Enum):
    """Enumeration of enrichment stages for progress tracking."""

    METADATA_FETCH = "metadata_fetch"
    TABLE_ENRICHMENT = "table_enrichment"
    MODEL_DESCRIPTION = "model_description"
    MODEL_METRICS = "model_metrics"
    VERIFIED_QUERIES = "verified_queries"
    COMPLETE = "complete"


@dataclass
class ProgressUpdate:
    """Represents a progress update during enrichment."""

    stage: EnrichmentStage
    current_step: int
    total_steps: int
    table_name: Optional[str] = None
    message: str = ""
    percentage: float = 0.0
    details: Optional[Dict[str, Any]] = None


class EnrichmentProgressTracker:
    """
    Tracks and reports progress during semantic model enrichment.

    This class manages progress reporting across multiple enrichment stages,
    calculating overall progress percentage and providing structured updates
    to UI components via callback functions.
    """

    def __init__(self, progress_callback: Optional[Callable[[ProgressUpdate], None]] = None):
        """
        Initialize the progress tracker.

        Args:
            progress_callback: Optional callback function to receive progress updates.
        """
        self.progress_callback = progress_callback
        self.current_stage = EnrichmentStage.METADATA_FETCH

        # Weight distribution across stages (should sum to 1.0)
        self.stage_weights = {
            EnrichmentStage.METADATA_FETCH: 0.05,      # 5% - Quick metadata collection
            EnrichmentStage.TABLE_ENRICHMENT: 0.70,    # 70% - Most time-consuming (multiple LLM calls)
            EnrichmentStage.MODEL_DESCRIPTION: 0.05,   # 5% - Single LLM call
            EnrichmentStage.MODEL_METRICS: 0.10,       # 10% - Single LLM call
            EnrichmentStage.VERIFIED_QUERIES: 0.10,    # 10% - Single LLM call + validation
        }

        # Track accumulated progress from completed stages
        self.completed_stage_progress = 0.0

    def update_progress(
        self,
        stage: EnrichmentStage,
        current: int,
        total: int,
        table_name: Optional[str] = None,
        message: str = "",
        details: Optional[Dict[str, Any]] = None
    ) -> None:
        """
        Update progress for the current enrichment stage.

        Args:
            stage: Current enrichment stage
            current: Current step number (0-based or 1-based)
            total: Total number of steps in this stage
            table_name: Optional table name being processed
            message: Human-readable progress message
            details: Optional additional progress details
        """
        # Update current stage if it has changed
        if self.current_stage != stage:
            self._stage_completed(self.current_stage)
            self.current_stage = stage

        # Calculate overall progress percentage
        percentage = self._calculate_overall_percentage(stage, current, total)

        # Create progress update
        update = ProgressUpdate(
            stage=stage,
            current_step=current,
            total_steps=total,
            table_name=table_name,
            message=message,
            percentage=percentage,
            details=details or {}
        )

        # Send update via callback
        if self.progress_callback:
            try:
                self.progress_callback(update)
            except Exception:
                # Silently ignore callback failures to avoid breaking enrichment
                pass

    def _stage_completed(self, stage: EnrichmentStage) -> None:
        """Mark a stage as completed and update accumulated progress."""
        if stage in self.stage_weights:
            self.completed_stage_progress += self.stage_weights[stage]

    def _calculate_overall_percentage(
        self,
        stage: EnrichmentStage,
        current: int,
        total: int
    ) -> float:
        """
        Calculate overall progress percentage across all stages.

        Args:
            stage: Current stage
            current: Current step in stage
            total: Total steps in stage

        Returns:
            Overall progress percentage (0.0 to 100.0)
        """
        # Get weight for current stage
        stage_weight = self.stage_weights.get(stage, 0.0)

        # Calculate progress within current stage
        if total > 0:
            stage_progress = min(current / total, 1.0) * stage_weight
        else:
            stage_progress = 0.0

        # Total progress = completed stages + current stage progress
        total_progress = self.completed_stage_progress + stage_progress

        # Convert to percentage and ensure bounds
        return min(max(total_progress * 100.0, 0.0), 100.0)

    def mark_complete(self) -> None:
        """Mark the entire enrichment process as complete."""
        self.update_progress(
            stage=EnrichmentStage.COMPLETE,
            current=1,
            total=1,
            message="Enrichment complete"
        )


def create_ui_progress_callback() -> Callable[[ProgressUpdate], None]:
    """
    Create a progress callback suitable for Streamlit UI updates.

    Returns:
        Callback function that formats progress updates for UI display.
    """
    def callback(update: ProgressUpdate) -> None:
        """Format and display progress update in UI."""
        # Build progress message
        stage_labels = {
            EnrichmentStage.METADATA_FETCH: "Collecting metadata",
            EnrichmentStage.TABLE_ENRICHMENT: "Enriching tables",
            EnrichmentStage.MODEL_DESCRIPTION: "Generating model description",
            EnrichmentStage.MODEL_METRICS: "Generating model metrics",
            EnrichmentStage.VERIFIED_QUERIES: "Generating verified queries",
            EnrichmentStage.COMPLETE: "Complete"
        }

        stage_label = stage_labels.get(update.stage, update.stage.value)

        if update.table_name:
            base_message = f"{stage_label} - {update.table_name}"
        else:
            base_message = stage_label

        if update.total_steps > 1:
            base_message += f" ({update.current_step}/{update.total_steps})"

        if update.message:
            full_message = f"{base_message}: {update.message}"
        else:
            full_message = base_message

        # In a real implementation, this would update the Streamlit UI
        # For now, we'll use the existing progress callback mechanism
        print(f"[{update.percentage:.1f}%] {full_message}")

    return callback


from __future__ import annotations

import json
import os
from dataclasses import dataclass
from http import HTTPStatus
from typing import Any, Dict, List, Optional
from urllib.parse import urlparse, urlunparse

from loguru import logger
try:
    import dashscope  # type: ignore
    from dashscope import Generation  # type: ignore
except ImportError:  # pragma: no cover - optional dependency
    dashscope = None  # type: ignore
    Generation = None  # type: ignore


@dataclass(frozen=True)
class DashscopeSettings:
    api_key: str
    model: str
    base_url: str = ""
    temperature: float = 0.2
    top_p: float = 0.85
    max_output_tokens: int = 512
    timeout_seconds: float = 45.0


class DashscopeError(RuntimeError):
    """Raised when DashScope returns an error response."""


@dataclass(frozen=True)
class DashscopeResponse:
    content: str
    request_id: Optional[str] = None


def _normalize_base_url(raw: str) -> str:
    raw = (raw or "").strip()
    if not raw:
        return ""
    parsed = urlparse(raw if "://" in raw else f"https://{raw}")
    if not parsed.netloc:
        return ""
    scheme = parsed.scheme or "https"
    netloc = parsed.netloc
    path = parsed.path or ""
    if "compatible-mode" in path:
        path = path.replace("compatible-mode", "api")
    if not path or path == "/":
        path = "/api/v1"
    if not path.endswith("/api/v1"):
        path = path.rstrip("/") + "/api/v1"
    return urlunparse((scheme, netloc, path, "", "", ""))


class DashscopeClient:
    def __init__(self, settings: DashscopeSettings) -> None:
        if dashscope is None or Generation is None:
            raise DashscopeError(
                "DashScope Python SDK is not installed. Please add `dashscope` to your environment."
            )
        self._settings = settings
        self._normalized_base_url = _normalize_base_url(settings.base_url)

    def chat_completion(self, messages: List[Dict[str, str]]) -> DashscopeResponse:
        if dashscope is None or Generation is None:  # pragma: no cover - double guard
            raise DashscopeError(
                "DashScope Python SDK is not installed. Please add `dashscope` to your environment."
            )

        dashscope.api_key = self._settings.api_key
        os.environ.setdefault("DASHSCOPE_API_KEY", self._settings.api_key)

        response = None
        managed_env_keys = [
            "DASHSCOPE_API_BASE",
            "DASHSCOPE_BASE_URL",
            "DASHSCOPE_COMPAT_URL",
        ]
        previous_env: Dict[str, Optional[str]] = {}
        try:
            for key in managed_env_keys:
                previous_env[key] = os.environ.pop(key, None)

            normalized_base_url = self._normalized_base_url
            if normalized_base_url:
                os.environ["DASHSCOPE_BASE_URL"] = normalized_base_url
                if hasattr(dashscope, "base_http_api_url"):
                    dashscope.base_http_api_url = normalized_base_url  # type: ignore[attr-defined]

            model_name = self._settings.model or "qwen-plus"
            dashscope_model: Any = model_name
            if isinstance(model_name, str):
                normalized = model_name.strip().lower().replace("-", "_")
                if normalized.endswith("_latest"):
                    normalized = normalized[: -len("_latest")]
                if "embedding" in normalized:
                    normalized = "qwen_plus"
                if hasattr(Generation.Models, normalized):
                    dashscope_model = getattr(Generation.Models, normalized)
                elif "qwen" in normalized:
                    dashscope_model = "qwen-plus"
                else:
                    dashscope_model = Generation.Models.qwen_plus

            response = Generation.call(
                model=dashscope_model,
                messages=messages,
                stream=False,
                result_format="message",
                temperature=self._settings.temperature,
                top_p=self._settings.top_p,
                max_output_tokens=self._settings.max_output_tokens,
                timeout=self._settings.timeout_seconds,
            )
        except Exception as exc:  # pragma: no cover - SDK raised error
            raise DashscopeError(f"DashScope request failed: {exc}") from exc
        finally:
            for key, value in previous_env.items():
                if value is None:
                    os.environ.pop(key, None)
                else:
                    os.environ[key] = value

        if response is None:
            raise DashscopeError("DashScope call did not return a response.")

        status = getattr(response, "status_code", None)
        if status != HTTPStatus.OK:
            error_message = getattr(response, "message", "Unknown error")
            error_code = getattr(response, "code", None)
            logger.error(
                "DashScope returned error ({}, code={}): {}",
                status,
                error_code,
                error_message,
            )
            raise DashscopeError(
                f"DashScope error {status} (code={error_code}): {error_message}"
            )

        output = getattr(response, "output", None)
        if not output or not hasattr(output, "choices"):
            raise DashscopeError(f"DashScope response missing output choices: {response}")

        choices = getattr(output, "choices")
        if not choices:
            raise DashscopeError("DashScope response contained no choices.")

        first = choices[0]
        message = getattr(first, "message", None)
        if not message or not hasattr(message, "content"):
            raise DashscopeError("DashScope response missing message content.")

        content = getattr(message, "content")
        if not content:
            raise DashscopeError("DashScope response returned empty content.")

        request_id = getattr(response, "request_id", None)
        return DashscopeResponse(content=content, request_id=request_id)



# TODO: Add tests for quoted columns, which are not well tested today.

import copy
from typing import List, Optional

import sqlglot
import sqlglot.expressions
from loguru import logger
from sqlglot import Dialect

from semantic_model_generator.protos import semantic_model_pb2
from semantic_model_generator.clickzetta_utils.clickzetta_connector import (
    OBJECT_DATATYPES,
)

_SQLGLOT_CLICKZETTA_KEY = "".join(["snow", "flake"])
ClickzettaDialect = Dialect.get_or_raise(_SQLGLOT_CLICKZETTA_KEY)

_LOGICAL_TABLE_PREFIX = "__"


def is_logical_table(table_name: str) -> bool:
    """Returns true if 'table_name' is a logical table name."""
    return table_name.startswith(_LOGICAL_TABLE_PREFIX) and len(table_name) > len(
        _LOGICAL_TABLE_PREFIX
    )


def logical_table_name(table: semantic_model_pb2.Table) -> str:
    """Returns the name of logical table for a given table.  E.g. __fact"""
    return _LOGICAL_TABLE_PREFIX + table.name  # type: ignore[no-any-return]


def fully_qualified_table_name(table: semantic_model_pb2.FullyQualifiedTable) -> str:
    """Returns fully qualified table name such as my_db.my_schema.my_table"""
    fqn = table.table
    if len(table.schema) > 0:
        fqn = f"{table.schema}.{fqn}"
    if len(table.database) > 0:
        fqn = f"{table.database}.{fqn}"
    return fqn  # type: ignore[no-any-return]


def is_aggregation_expr(col: semantic_model_pb2.Column) -> bool:
    """Check if an expr contains aggregation function.
    Note: only flag True for aggregations that would changes number of rows of data.
    For window function, given the operation will produce value per row, mark as False here.

    Raises:
        ValueError: if expr is not parsable, or if aggregation expressions in non-measure columns.
    """
    parsed = sqlglot.parse_one(col.expr, dialect=ClickzettaDialect)
    agg_func = list(parsed.find_all(sqlglot.expressions.AggFunc))
    window = list(parsed.find_all(sqlglot.expressions.Window))
    # We've confirmed window functions cannot appear inside aggregate functions
    # (gets execution error msg: Window function [SUM(...) OVER (PARTITION BY ...)] may not appear inside an aggregate function).
    # So if there's a window function present there can't also be an aggregate function applied to the window function.
    if len(agg_func) > 0 and len(window) == 0:
        if col.kind != 2:
            raise ValueError("Only allow aggregation expressions for measures.")
        return True
    return False


def _is_physical_table_column(col: semantic_model_pb2.Column) -> bool:
    """Returns whether the column refers to a single raw table column."""
    try:
        parsed = sqlglot.parse_one(col.expr, dialect=ClickzettaDialect)
        return isinstance(parsed, sqlglot.expressions.Column)
    except Exception as ex:
        logger.warning(
            f"Failed to parse sql expression: {col.expr}. Error: {ex}. {col}"
        )
        return False


def _is_identifier_quoted(col_name: str) -> bool:
    return '"' in col_name


def remove_ltable_cte(sql_w_ltable_cte: str, table_names: list[str]) -> str:
    """
    Given a SQL with prefix'd logical table conversion CTE(s), remove the logical table conversions.
    Args:
        sql_w_ltable_cte: the sql with logical table conversion CTE(s).
        table_names: list of tables in the semantic model.

    Returns: the sql without the logical table conversion CTE.
    Raises: ValueError if didn't find any CTE or parsed first CTE is not logical table CTE.
    """
    ast = sqlglot.parse_one(sql_w_ltable_cte, read=ClickzettaDialect)
    with_ = ast.args.get("with")
    if with_ is None:
        raise ValueError("Analyst queries must contain the logical CTE.")
    if not is_logical_table(with_.expressions[0].alias):
        raise ValueError("Analyst queries must contain the logical CTE.")

    table_names_lower = [table_name.lower() for table_name in table_names]
    # Iterate through all CTEs, and filter out logical table CTEs.
    # This is done by checking if the CTE alias starts with the logical table prefix and if the alias is in a table in the semantic model.
    non_logical_cte = [
        cte
        for cte in with_.expressions
        if not is_logical_table(cte.alias)
        or cte.alias.replace(_LOGICAL_TABLE_PREFIX, "").lower() not in table_names_lower
    ]

    # Replace the original expressions list with the filtered list
    with_.set("expressions", non_logical_cte)

    # If no expressions are left for whatever reason, remove the entire WITH clause.
    if not with_.expressions:
        ast.set("with", None)

    sql_without_logical_cte = ast.sql(dialect=ClickzettaDialect, pretty=True)
    return sql_without_logical_cte  # type: ignore [no-any-return]


def _validate_col(column: semantic_model_pb2.Column) -> None:
    if " " in column.name.strip():
        raise ValueError(
            f"Please do not include spaces in your column name: {column.name}"
        )
    if column.data_type.upper() in OBJECT_DATATYPES:
        raise ValueError(
            f"We do not support object datatypes in the semantic model. Col {column.name} has data type {column.data_type}. Please remove this column from your semantic model or flatten it to non-object type."
        )


def validate_all_cols(table: semantic_model_pb2.Table) -> None:
    for column in table.columns:
        _validate_col(column)


def _get_col_expr(column: semantic_model_pb2.Column) -> str:
    """Return column expr in SQL format.
    Raise errors if columns is of OBJECT_DATATYPES, which we do not support today."""
    return (
        f"{column.expr.strip()} as {column.name.strip()}"
        if column.expr.strip().lower() != column.name.strip().lower()
        else f"{column.expr.strip()}"
    )


def _generate_cte_for(
    table: semantic_model_pb2.Table, columns: List[semantic_model_pb2.Column]
) -> str:
    """
    Returns a CTE representing a logical table that selects 'col' columns from 'table'.
    """

    if len(columns) == 0:
        raise ValueError("Please include at least one column to generate CTE on.")
    else:
        expr_columns = [_get_col_expr(col) for col in columns]
        cte = f"WITH {logical_table_name(table)} AS (\n"
        cte += "SELECT \n"
        cte += ",\n".join(expr_columns) + "\n"
        cte += f"FROM {fully_qualified_table_name(table.base_table)}"
        cte += ")"
        return cte


def get_all_physical_column_references(
    column: semantic_model_pb2.Column,
) -> List[str]:
    """Returns a set of column names referenced in the column expression.

    For example, the following column expressions yield the following return values:
    foo -> [foo]
    foo+bar -> [foo, bar]
    sum(foo) -> [foo]
    """
    try:
        parsed = sqlglot.parse_one(column.expr, dialect=ClickzettaDialect)
        col_names = set()
        for col in parsed.find_all(sqlglot.expressions.Column):
            # TODO(renee): Handle quoted columns.
            col_name = col.name.lower()
            if col.this.quoted:
                col_name = col.name
            col_names.add(col_name)
        return sorted(list(col_names))
    except Exception as ex:
        raise ValueError(f"Failed to parse sql expression: {column.expr}. Error: {ex}")


def direct_mapping_logical_columns(
    table: semantic_model_pb2.Table,
) -> List[semantic_model_pb2.Column]:
    """
    Returns a list of logical columns that map 1:1 to an underlying physical column
    (i.e. logical table's expression is simply the physical column name) in this table.
    """
    ret: List[semantic_model_pb2.Column] = []
    for c in table.columns:
        if _is_physical_table_column(c):
            ret.append(c)
    return ret


def _enrich_column_in_expr_with_aggregation(
    table: semantic_model_pb2.Table,
) -> semantic_model_pb2.Table:
    """
    Expands the logical columns of 'table' to include columns mentioned in a logical columns
    with an aggregate expression. E.g. for a logical column called CPC with expr sum(cost) / sum(clicks),
    adds logical columns for "cost" and "clicks", if not present.
    """
    direct_mapping_lcols = [
        c.name.lower() for c in direct_mapping_logical_columns(table)
    ]
    cols_to_append = set()
    for col in table.columns:
        if not is_aggregation_expr(col):
            continue
        for pcol in get_all_physical_column_references(col):
            # If the physical column doesn't have a direct mapping logical column
            # with the same name, then we need to add a new logical column for it.
            # Note that this may introduce multiple logical columns directly referencing
            # the same physical column, something we should improve up, perhaps by
            # rewriting the expression to use existing direct mapping logical columns
            # whenever preset.
            if pcol not in direct_mapping_lcols:
                cols_to_append.add(pcol)

    original_cols = {col.name.lower(): col.expr for col in table.columns}
    ret = copy.deepcopy(table)
    # Insert in sorted order to make this method deterministic.
    for c in sorted(cols_to_append):
        if c in original_cols:
            logger.warning(
                f"Not adding a logical column for physical column {c} in table {table.name}, "
                f"since this logical column already exists with expression {original_cols[c]}"
            )
        else:
            new_col = semantic_model_pb2.Column(name=c, expr=c)
            ret.columns.append(new_col)
    return ret


def _generate_non_agg_cte(table: semantic_model_pb2.Table) -> Optional[str]:
    """
    Returns a CTE representing a logical table that selects 'col' columns from 'table' except for aggregation columns.
    """
    filtered_cols = [col for col in table.columns if not is_aggregation_expr(col)]
    if len(filtered_cols) > 0:
        return _generate_cte_for(table, filtered_cols)
    else:
        return None


def _convert_to_clickzetta_sql(sql: str) -> str:
    """
    Converts a given SQL statement to ClickZetta SQL syntax using SQLGlot.

    Args:
    sql (str): The SQL statement to convert.

    Returns:
    str: The SQL statement in ClickZetta syntax.
    """
    try:
        expression = sqlglot.parse_one(sql, dialect=ClickzettaDialect)
    except Exception as e:
        raise ValueError(
            f"Unable to parse sql statement.\n Provided sql: {sql}\n. Error: {e}"
        )

    return expression.sql(dialect=ClickzettaDialect)


def generate_select(
    table_in_column_format: semantic_model_pb2.Table, limit: int
) -> List[str]:
    """Generate select query for all columns for validation purpose."""
    sqls_to_return: List[str] = []
    # Generate select query for columns without aggregation exprs.
    non_agg_cte = _generate_non_agg_cte(table_in_column_format)
    if non_agg_cte is not None:
        non_agg_sql = (
            non_agg_cte
            + f"SELECT * FROM {logical_table_name(table_in_column_format)} LIMIT {limit}"
        )
        sqls_to_return.append(_convert_to_clickzetta_sql(non_agg_sql))

    # Generate select query for columns with aggregation exprs.
    agg_cols = [
        col for col in table_in_column_format.columns if is_aggregation_expr(col)
    ]
    if len(agg_cols) == 0:
        return sqls_to_return
    else:
        agg_cte = _generate_cte_for(table_in_column_format, agg_cols)
        agg_sql = (
            agg_cte
            + f"SELECT * FROM {logical_table_name(table_in_column_format)} LIMIT {limit}"
        )
        sqls_to_return.append(_convert_to_clickzetta_sql(agg_sql))
    return sqls_to_return


def expand_all_logical_tables_as_ctes(
    sql_query: str, model_in_column_format: semantic_model_pb2.SemanticModel
) -> str:
    """
    Returns a SQL query that expands all logical tables contained in ctx as ctes.
    """

    def generate_full_logical_table_ctes(
        ctx: semantic_model_pb2.SemanticModel,
    ) -> List[str]:
        """
        Given an arbitrary SQL, returns a list of CTEs representing all the logical tables
        referenced in it.
        """
        ctes: List[str] = []
        for table in ctx.tables:
            # Append all columns and expressions for the logical table.
            # If table contains expr with aggregations, enrich its referred columns into the table.
            table_ = _enrich_column_in_expr_with_aggregation(table)
            cte = _generate_non_agg_cte(table_)
            if cte is not None:
                ctes.append(cte)
        return ctes

    # Step 1: Generate a CTE for each logical table referenced in the query.
    ctes = generate_full_logical_table_ctes(model_in_column_format)

    # Step 2: Parse each generated CTE as a 'WITH' clause.
    new_withs = []
    for cte in ctes:
        new_withs.append(
            sqlglot.parse_one(
                cte, read=ClickzettaDialect, into=sqlglot.expressions.With
            )
        )

    # Step 3: Prefix the CTEs to the original query.
    ast = sqlglot.parse_one(sql_query, read=ClickzettaDialect)
    with_ = ast.args.get("with")
    # If the query doesn't have a WITH clause, then generate one.
    if with_ is None:
        merged_with = new_withs[0]
        remaining_ctes = [w.expressions[0] for w in new_withs[1:]]
        merged_with.set("expressions", merged_with.expressions + remaining_ctes)
        ast.set("with", merged_with)
    # If the query already has a WITH clause, prefix the CTEs to it.
    else:
        new_ctes = [w.expressions[0] for w in new_withs]
        with_.set("expressions", new_ctes + with_.expressions)
    return ast.sql(dialect=ClickzettaDialect, pretty=True)  # type: ignore [no-any-return]


def context_to_column_format(
    ctx: semantic_model_pb2.SemanticModel,
) -> semantic_model_pb2.SemanticModel:
    """
    Converts semantic_model_pb2.SemanticModel from a dimension/measure format to a column format.
    Returns a new semantic_model_pb2.SemanticModel object that's in column format.
    """
    ret = semantic_model_pb2.SemanticModel()
    ret.CopyFrom(ctx)
    for table in ret.tables:
        column_format = len(table.columns) > 0
        dimension_measure_format = (
            len(table.dimensions) > 0
            or len(table.time_dimensions) > 0
            or len(table.measures) > 0
        )
        if column_format and dimension_measure_format:
            raise ValueError(
                "table {table.name} defines both columns and dimensions/time_dimensions/measures."
            )
        if column_format:
            continue
        for d in table.dimensions:
            col = semantic_model_pb2.Column()
            col.kind = semantic_model_pb2.ColumnKind.dimension
            col.name = d.name
            col.synonyms.extend(d.synonyms)
            col.description = d.description
            col.expr = d.expr
            col.data_type = d.data_type
            col.unique = d.unique
            col.sample_values.extend(d.sample_values)
            table.columns.append(col)
        del table.dimensions[:]

        for td in table.time_dimensions:
            col = semantic_model_pb2.Column()
            col.kind = semantic_model_pb2.ColumnKind.time_dimension
            col.name = td.name
            col.synonyms.extend(td.synonyms)
            col.description = td.description
            col.expr = td.expr
            col.data_type = td.data_type
            col.unique = td.unique
            col.sample_values.extend(td.sample_values)
            table.columns.append(col)
        del table.time_dimensions[:]

        for m in table.measures:
            col = semantic_model_pb2.Column()
            col.kind = semantic_model_pb2.ColumnKind.measure
            col.name = m.name
            col.synonyms.extend(m.synonyms)
            col.description = m.description
            col.expr = m.expr
            col.data_type = m.data_type
            col.default_aggregation = m.default_aggregation
            col.sample_values.extend(m.sample_values)
            table.columns.append(col)
        del table.measures[:]
    return ret
